#!/usr/bin/env python
# -*- coding: utf-8 -*-
u"""
Multipletau provides a multiple-τ algorithm Python 3.x with
:py:mod:`numpy` as its sole dependency.

Multiple-τ correlation is computed on a logarithmic scale (less
data points are computed) and is thus much faster than conventional
correlation on a linear scale such as  :py:func:`numpy.correlate`.

Recommended literature
----------------------

- Klaus Schaetzel and Rainer Peters; *Noise on multiple-tau photon
  correlation data*. Proc. SPIE 1430, Photon Correlation
  Spectroscopy: Multicomponent Systems, 109 (June 1, 1991);
  http://doi.org/10.1117/12.44160

- Thorsten Wohland, Rudolf Rigler, and Horst Vogel; *The Standard
  Deviation in Fluorescence Correlation Spectroscopy*. Biophysical
  Journal, 80 (June 1, 2001);
  http://dx.doi.org/10.1016/S0006-3495(01)76264-9

Obtaining multipletau
---------------------
If you have Python and :py:mod:`numpy` installed, simply run::

   pip install multipletau

The source code of multipletau is available at
https://github.com/FCS-analysis/multipletau.


Citing multipletau
------------------
The multipletau package should be cited like this (replace "x.x.x"
with the actual version of multipletau used and "DD Month YYYY"
with a matching date).

.. topic:: cite

    Paul Müller (2012) *Python multiple-tau algorithm* (Version x.x.x)
    [Computer program].
    Available at https://pypi.python.org/pypi/multipletau/
    (Accessed DD Month YYYY)


You can find out what version you are using by typing
(in a Python console):


    >>> import multipletau
    >>> multipletau.__version__
    '0.3.0'


Usage
-----
The package is straightforward to use. Here is a quick example:


    >>> import numpy as np
    >>> import multipletau
    >>> a = np.linspace(2,5,42)
    >>> v = np.linspace(1,6,42)
    >>> multipletau.correlate(a, v, m=2)
    array([[   0.        ,  569.56097561],
           [   1.        ,  549.87804878],
           [   2.        ,  530.37477692],
           [   4.        ,  491.85812017],
           [   8.        ,  386.39500297]])

"""
from .core import autocorrelate, correlate, correlate_numpy  # noqa: F401
from ._version import __version__, __version_tuple__  # noqa: F401

__author__ = u"Paul Müller"
__license__ = "BSD (3 clause)"



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Methods for correlated noise generation"""

from __future__ import division
from __future__ import print_function

import numpy as np

__all__ = ["noise_exponential", "noise_cross_exponential"]


def noise_exponential(N, tau=20, variance=1, deltat=1):
    """Generate exponentially correlated noise.

    Parameters
    ----------
    N : integer
       Total number of samples
    tau : float
       Correlation time of the exponential in `deltat`
    variance : float
       Variance of the noise
    deltat : float
       Bin size of output array, defines the time scale of `tau`

    Returns
    -------
    a : ndarray
       Exponentially correlated noise.
    """
    # time constant (inverse of correlationtime tau)
    g = deltat / tau
    # variance
    s0 = variance

    # normalization factor (memory of the trace)
    exp_g = np.exp(-g)
    one_exp_g = 1 - exp_g
    z_norm_factor = np.sqrt(1 - np.exp(-2 * g)) / one_exp_g

    # create random number array
    # generates random numbers in interval [0,1)
    randarray = np.random.random(N)
    # make numbers random in interval [-1,1)
    randarray = 2 * (randarray - 0.5)

    # simulate exponential random behavior
    a = np.zeros(N)
    a[0] = one_exp_g * randarray[0]
    for i in np.arange(N - 1) + 1:
        a[i] = exp_g * a[i - 1] + one_exp_g * randarray[i]

        # Solving the equation iteratively leads to this equation:
        # j = np.arange(i)
        # a[i] = a[0]*exp_g**(i) + \
        #       one_exp_g)*np.sum(exp_g**(i-1-j)*randarray[1:i+1])

    a = a * z_norm_factor * s0
    return a


def noise_cross_exponential(N, tau=20, variance=1, deltat=1):
    """Generate exponentially cross-correlated noise.

    Parameters
    ----------
    N : integer
       Total number of samples
    tau : float
       Correlation time of the exponential in `deltat`
    variance : float
       Variance of the noise
    deltat : float
       Bin size of output array, defines the time scale of `tau`

    Returns
    -------
    a, randarray : ndarrays
       Array `a` has positive exponential correlation to the 'truly'
       random array `randarray`.
    """
    # time constant (inverse of correlationtime tau)
    g = deltat / tau
    # variance
    s0 = variance
    # normalization factor (memory of the trace)
    exp_g = np.exp(-g)
    one_exp_g = 1 - exp_g
    z_norm_factor = np.sqrt(1 - np.exp(-2 * g)) / one_exp_g

    # create random number array
    # generates random numbers in interval [0,1)
    randarray = np.random.random(N)
    # make numbers random in interval [-1,1)
    randarray = 2 * (randarray - 0.5)

    # simulate exponential random behavior
    a = np.zeros(N)
    a[0] = one_exp_g * randarray[0]

    b = np.zeros(N)
    b[0] = one_exp_g * randarray[0]
    # slow
    # for i in np.arange(N-1)+1:
    #    for j in np.arange(i-1):
    #        a[i] += exp_g**j*randarray[i-j]
    #    a[i] += one_exp_g*randarray[i]
    # faster
    j = np.arange(N + 5)
    for i in np.arange(N - 1) + 1:
        a[i] += np.sum(exp_g**j[2:i + 1] * randarray[2:i + 1][::-1])
        a[i] += one_exp_g * randarray[i]

    a *= z_norm_factor * s0
    randarray = randarray * z_norm_factor * s0

    return a, randarray



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests autocorrelation algorithm"""
from __future__ import division, print_function

import os
import sys

import numpy as np
import pytest

import multipletau


def get_reference_data(funcname, pyfile):
    adir = os.path.dirname(pyfile)+"/data/"
    aname = os.path.basename(pyfile)+"_"+funcname+".npy"
    return np.load(adir + aname)


def get_sample_arrays():
    a = [-4.3,   1,    9, -99.2, 13]
    b = [9921, 281, 23.5,   5.3, 77]
    ll = [33,  92,   47,    54, 99]
    r = [0,   1,   12,     4,  0]
    p = [1,   4,   .5,     2,  3]
    arrs = []

    for ai, bi, li, ri, pi in zip(a, b, ll, r, p):
        x = np.linspace(ai, bi, li)
        arr = (x*np.roll(x, ri))**pi
        arrs.append(arr)

    return arrs


def test_ac_copy():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    res1 = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=True,
                                      copy=True,
                                      dtype=np.float64)
        res1.append(r)

    res2 = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=True,
                                      copy=False,
                                      dtype=np.float64)
        res2.append(r)

    # simple test if result is the same
    assert np.all(np.concatenate(res1) == np.concatenate(res2))

    arrs = np.concatenate(arrs)
    refarrs = np.concatenate(get_sample_arrays())

    # make sure the copy function really changes something
    assert not np.all(arrs == refarrs)


@pytest.mark.filterwarnings("ignore::multipletau.core.DtypeWarning")
def test_ac_dtype():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.round(get_sample_arrays()[0])

    # integer
    rf = multipletau.autocorrelate(a=a,
                                   m=16,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)

    ri = multipletau.autocorrelate(a=a,
                                   m=16,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.uint)

    ri2 = multipletau.autocorrelate(a=np.array(a, dtype=np.uint),
                                    m=16,
                                    deltat=1,
                                    normalize=True,
                                    copy=True,
                                    dtype=None)

    assert ri.dtype == np.dtype(
        np.float64), "if wrong dtype, dtype should default to np.float64"
    assert ri2.dtype == np.dtype(
        np.float64), "if wrong dtype, dtype should default to np.float64"
    assert np.all(
        rf == ri), "result should be the same, because input us the same"
    assert np.all(
        rf == ri2), "result should be the same, because input us the same"


def test_ac_m():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    ms = [8, 16, 32, 64, 128]
    a = np.concatenate(arrs)

    res = []
    for m in ms:
        r = multipletau.autocorrelate(a=a,
                                      m=m,
                                      deltat=1,
                                      normalize=False,
                                      copy=True,
                                      dtype=np.float64)
        res.append(r)

        # test minimal length of array
        multipletau.autocorrelate(a=a[:2*m],
                                  m=m,
                                  deltat=1,
                                  normalize=False,
                                  copy=True,
                                  dtype=np.float64)

    res = np.concatenate(res)
    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-15)


@pytest.mark.filterwarnings("ignore::multipletau.core.InvalidMWarning")
def test_ac_m_wrong():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = get_sample_arrays()[0]

    # integer
    r1 = multipletau.autocorrelate(a=a,
                                   m=16,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)

    r2 = multipletau.autocorrelate(a=a,
                                   m=15,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)

    r3 = multipletau.autocorrelate(a=a,
                                   m=15.5,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)

    r4 = multipletau.autocorrelate(a=a,
                                   m=14.5,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)

    r5 = multipletau.autocorrelate(a=a,
                                   m=16.,
                                   deltat=1,
                                   normalize=True,
                                   copy=True,
                                   dtype=np.float64)
    assert np.all(r1 == r2)
    assert np.all(r1 == r3)
    assert np.all(r1 == r4)
    assert np.all(r1 == r5)


def test_ac_normalize():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    res = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=True,
                                      copy=True,
                                      dtype=np.float64)
        res.append(r)

    res = np.concatenate(res)
    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-14)


def test_ac_simple():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    res = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=False,
                                      copy=True,
                                      dtype=np.float64)
        res.append(r)

    res = np.concatenate(res)
    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-15)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""test returning exact sum plus normalization factor"""
import numpy as np

from multipletau import autocorrelate, correlate


def test_ac_return_sum():
    ist, ist_count = autocorrelate(range(42), m=2, dtype=np.float64,
                                   ret_sum=True)
    soll = np.array([[0.000000e+00, 2.382100e+04],
                     [1.000000e+00, 2.296000e+04],
                     [2.000000e+00, 2.210000e+04],
                     [4.000000e+00, 1.018875e+04],
                     [8.000000e+00, 3.586000e+03]])
    soll_count = [42., 41., 40., 19.,  8.]
    assert np.allclose(soll, ist)
    assert np.allclose(soll_count, ist_count)


def test_cc_compress_average():
    ist, ist_count = correlate(range(42), range(1, 43), m=2, dtype=np.float64,
                               ret_sum=True)
    soll = np.array([[0.000000e+00, 2.468200e+04],
                     [1.000000e+00, 2.382100e+04],
                     [2.000000e+00, 2.296000e+04],
                     [4.000000e+00, 1.061625e+04],
                     [8.000000e+00, 3.774000e+03]])
    soll_count = [42., 41., 40., 19.,  8.]
    assert np.allclose(soll, ist)
    assert np.allclose(soll_count, ist_count)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Compare to numpy data"""
import sys

import numpy as np

import multipletau

from test_correlate import get_sample_arrays_cplx


def test_corresponds_ac():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.concatenate(get_sample_arrays_cplx()).real
    m = 16

    restau = multipletau.autocorrelate(a=1*a,
                                       m=m,
                                       copy=True,
                                       normalize=True,
                                       dtype=np.float64)

    reslin = multipletau.correlate_numpy(a=1*a,
                                         v=1*a,
                                         copy=True,
                                         normalize=True,
                                         dtype=np.float64)

    idx = np.array(restau[:, 0].real, dtype=int)[:m]

    assert np.allclose(reslin[idx, 1], restau[:m, 1])


def test_corresponds_ac_first_loop():
    """
    numpy correlation:
    G_m = sum_i(a_i*a_{i+m})

    multipletau correlation 2nd order:
    b_j = (a_{2i} + a_{2i+1} / 2)
    G_m = sum_j(b_j*b_{j+1})
        = 1/4*sum_i(a_{2i}   * a_{2i+m}   +
                    a_{2i}   * a_{2i+m+1} +
                    a_{2i+1} * a_{2i+m}   +
                    a_{2i+1} * a_{2i+m+1}
                    )

    The values after the first m+1 lag times in the multipletau
    correlation differ from the normal correlation, because the
    traces are averaged over two consecutive items, effectively
    halving the size of the trace. The multiple-tau correlation
    can be compared to the regular correlation by using an even
    sized sequence (here 222) in which the elements 2i and 2i+1
    are equal, as is done in this test.
    """
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = [arr / np.average(arr) for arr in get_sample_arrays_cplx()]
    a = np.concatenate(a)[:222]
    # two consecutive elements are the same, so the multiple-tau method
    # corresponds to the numpy correlation for the first loop.
    a[::2] = a[1::2]

    for m in [2, 4, 6, 8, 10, 12, 14, 16]:
        restau = multipletau.correlate(a=a,
                                       v=a.imag+1j*a.real,
                                       m=m,
                                       copy=True,
                                       normalize=False,
                                       dtype=np.complex128)

        reslin = multipletau.correlate_numpy(a=a,
                                             v=a.imag+1j*a.real,
                                             copy=True,
                                             normalize=False,
                                             dtype=np.complex128)

        idtau = np.where(restau[:, 0] == m+2)[0][0]
        tau3 = restau[idtau, 1]  # m+1 initial bins

        idref = np.where(reslin[:, 0] == m+2)[0][0]
        tau3ref = reslin[idref, 1]

        assert np.allclose(tau3, tau3ref)


def test_corresponds_ac_nonormalize():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.concatenate(get_sample_arrays_cplx()).real
    m = 16

    restau = multipletau.autocorrelate(a=1*a,
                                       m=m,
                                       copy=True,
                                       normalize=False,
                                       dtype=np.float64)

    reslin = multipletau.correlate_numpy(a=1*a,
                                         v=1*a,
                                         copy=True,
                                         normalize=False,
                                         dtype=np.float64)

    idx = np.array(restau[:, 0].real, dtype=int)[:m+1]

    assert np.allclose(reslin[idx, 1], restau[:m+1, 1])


def test_corresponds_cc():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.concatenate(get_sample_arrays_cplx())
    m = 16

    restau = multipletau.correlate(a=a,
                                   v=a.imag+1j*a.real,
                                   m=m,
                                   copy=True,
                                   normalize=True,
                                   dtype=np.complex128)

    reslin = multipletau.correlate_numpy(a=a,
                                         v=a.imag+1j*a.real,
                                         copy=True,
                                         normalize=True,
                                         dtype=np.complex128)

    idx = np.array(restau[:, 0].real, dtype=int)[:m+1]

    assert np.allclose(reslin[idx, 1], restau[:m+1, 1])


def test_corresponds_cc_nonormalize():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.concatenate(get_sample_arrays_cplx())
    m = 16

    restau = multipletau.correlate(a=a,
                                   v=a.imag+1j*a.real,
                                   m=m,
                                   copy=True,
                                   normalize=False,
                                   dtype=np.complex128)

    reslin = multipletau.correlate_numpy(a=a,
                                         v=a.imag+1j*a.real,
                                         copy=True,
                                         normalize=False,
                                         dtype=np.complex128)

    idx = np.array(restau[:, 0].real, dtype=int)[:m+1]

    assert np.allclose(reslin[idx, 1], restau[:m+1, 1])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test correlation-autocorrelation identity"""
from __future__ import division, print_function

import sys

import numpy as np

import multipletau

from test_autocorrelate import get_sample_arrays


def test_ac_cc_m():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    ms = [8, 16, 32, 64, 128]
    a = np.concatenate(arrs)

    res = []
    for m in ms:
        r = multipletau.autocorrelate(a=a,
                                      m=m,
                                      deltat=1,
                                      normalize=False,
                                      copy=True,
                                      dtype=np.float64)
        res.append(r)
    res = np.concatenate(res)

    rescc = []
    for m in ms:
        r = multipletau.correlate(a=a, v=a,
                                  m=m,
                                  deltat=1,
                                  normalize=False,
                                  copy=True,
                                  dtype=np.float64)
        rescc.append(r)
        # test minimal length of array
        multipletau.correlate(a=a[:2*m], v=a[:2*m],
                              m=m,
                              deltat=1,
                              normalize=False,
                              copy=True,
                              dtype=np.float64)

    rescc = np.concatenate(rescc)
    assert np.all(res == rescc)


def test_ac_cc_normalize():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    res = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=True,
                                      copy=True,
                                      dtype=np.float64)
        res.append(r)

    res = np.concatenate(res)

    rescc = []
    for a in arrs:
        r = multipletau.correlate(a=a, v=a,
                                  m=16,
                                  deltat=1,
                                  normalize=True,
                                  copy=True,
                                  dtype=np.float64)
        rescc.append(r)

    rescc = np.concatenate(rescc)

    assert np.all(res == rescc)


def test_ac_cc_simple():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays()

    rescc = []
    for a in arrs:
        r = multipletau.correlate(a=a, v=a,
                                  m=16,
                                  deltat=1,
                                  normalize=False,
                                  copy=True,
                                  dtype=np.float64)
        rescc.append(r)

    rescc = np.concatenate(rescc)

    resac = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=False,
                                      copy=True,
                                      dtype=np.float64)
        resac.append(r)

    resac = np.concatenate(resac)

    assert np.all(resac == rescc)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test correlation-autocorrelation identity"""
from __future__ import division, print_function

import numpy as np

import multipletau

from test_autocorrelate import get_sample_arrays


def test_ac():
    arrs = get_sample_arrays()

    try:
        multipletau.autocorrelate(a=arrs[0],
                                  copy=2)
    except ValueError as e:
        assert "`copy` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.autocorrelate(a=arrs[0],
                                  ret_sum=2)
    except ValueError as e:
        assert "`ret_sum` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.autocorrelate(a=arrs[0],
                                  normalize=2)
    except ValueError as e:
        assert "`normalize` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.autocorrelate(a=arrs[0],
                                  compress="peter")
    except ValueError as e:
        assert "Invalid value for `compress`!" in e.args[0]
    else:
        assert False

    try:
        multipletau.autocorrelate(a=arrs[0],
                                  normalize=True,
                                  ret_sum=True)
    except ValueError as e:
        assert "'normalize' and 'ret_sum' must not both be True!" in e.args
    else:
        assert False


def test_ac_trace0():
    arrs = get_sample_arrays()
    try:
        multipletau.autocorrelate(a=arrs[0] - np.mean(arrs[0]),
                                  normalize=True)
    except ValueError as e:
        assert "Cannot normalize: Average of `a` is zero!" in e.args
    else:
        assert False


def test_ac_tracesize():
    arrs = get_sample_arrays()
    try:
        multipletau.autocorrelate(a=arrs[0][:31],
                                  m=16)
    except ValueError as e:
        assert '`len(a)` must be >= `2m`!' in e.args
    else:
        assert False


def test_cc():
    arrs = get_sample_arrays()

    try:
        multipletau.correlate(a=arrs[0], v=arrs[0],
                              copy=2)
    except ValueError as e:
        assert "`copy` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.correlate(a=arrs[0], v=arrs[0],
                              ret_sum=2)
    except ValueError as e:
        assert "`ret_sum` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.correlate(a=arrs[0], v=arrs[0],
                              normalize=2)
    except ValueError as e:
        assert "`normalize` must be boolean!" in e.args
    else:
        assert False

    try:
        multipletau.correlate(a=arrs[0], v=arrs[0],
                              compress="peter")
    except ValueError as e:
        assert "Invalid value for `compress`!" in e.args[0]
    else:
        assert False

    try:
        multipletau.correlate(a=arrs[0], v=arrs[0],
                              normalize=True,
                              ret_sum=True)
    except ValueError as e:
        assert "'normalize' and 'ret_sum' must not both be True!" in e.args
    else:
        assert False


def test_cc_trace0():
    arrs = get_sample_arrays()
    try:
        multipletau.correlate(a=arrs[0] - np.mean(arrs[0]),
                              v=arrs[0],
                              normalize=True)
    except ValueError as e:
        assert "Cannot normalize: Average of `a` is zero!" in e.args
    else:
        assert False

    try:
        multipletau.correlate(a=arrs[0],
                              v=arrs[0] - np.mean(arrs[0]),
                              normalize=True)
    except ValueError as e:
        assert "Cannot normalize: Average of `v` is zero!" in e.args
    else:
        assert False


def test_cc_tracesize():
    arrs = get_sample_arrays()
    try:
        multipletau.correlate(a=arrs[0][:31],
                              v=arrs[0][:31],
                              m=16)
    except ValueError as e:
        assert '`len(a)` must be >= `2m`!' in e.args
    else:
        assert False


def test_cc_samesize():
    arrs = get_sample_arrays()
    try:
        multipletau.correlate(a=arrs[0],
                              v=arrs[1],
                              normalize=True)
    except ValueError as e:
        assert "`a` and `v` must have same length!" in e.args
    else:
        assert False


def test_numpy_cc_trace0():
    arrs = get_sample_arrays()
    try:
        multipletau.correlate_numpy(a=arrs[0] - np.mean(arrs[0]),
                                    v=arrs[0],
                                    normalize=True)
    except ValueError as e:
        assert "Cannot normalize: Average of `a` is zero!" in e.args
    else:
        assert False

    try:
        multipletau.correlate_numpy(a=arrs[0],
                                    v=arrs[0] - np.mean(arrs[0]),
                                    normalize=True)
    except ValueError as e:
        assert "Cannot normalize: Average of `v` is zero!" in e.args
    else:
        assert False


def test_numpy_cc_samesize():
    arrs = get_sample_arrays()
    try:
        multipletau.correlate_numpy(a=arrs[0],
                                    v=arrs[1],
                                    normalize=True)
    except ValueError as e:
        assert "`a` and `v` must have same length!" in e.args
    else:
        assert False


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Comparison of correlation methods

This example illustrates the differences between the
:py:mod:`multipletau` correlation methods
(:py:func:`multipletau.autocorrelate`,
:py:func:`multipletau.correlate`) and :py:func:`numpy.correlate`.

This example requires ``noise_generator.py`` to be present in the
current working directory.
"""
from matplotlib import pylab as plt
import numpy as np

from multipletau import autocorrelate, correlate, correlate_numpy

from noise_generator import noise_exponential, noise_cross_exponential


# starting parameters
N = np.int(np.pi * 1e3)
countrate = 250. * 1e-3  # in Hz
taudiff = 55.  # in us
deltat = 2e-6  # time discretization [s]
normalize = True

# time factor
taudiff *= deltat

# create noise for autocorrelation
data = noise_exponential(N, taudiff, deltat=deltat)
data -= np.average(data)
if normalize:
    data += countrate
# perform autocorrelation (multipletau)
gac_mt = autocorrelate(data, deltat=deltat, normalize=normalize)
# numpy.correlate for comparison
gac_np = correlate_numpy(data, data, deltat=deltat,
                         normalize=normalize)
# calculate model curve for autocorrelation
x = gac_np[:, 0]
amp = np.correlate(data - np.average(data), data - np.average(data),
                   mode="valid")
if normalize:
    amp /= len(data) * countrate**2
y = amp * np.exp(-x / taudiff)

# create noise for cross-correlation
a, v = noise_cross_exponential(N, taudiff, deltat=deltat)
a -= np.average(a)
v -= np.average(v)
if normalize:
    a += countrate
    v += countrate
gcc_forw_mt = correlate(a, v, deltat=deltat, normalize=normalize)  # forward
gcc_back_mt = correlate(v, a, deltat=deltat, normalize=normalize)  # backward
# numpy.correlate for comparison
gcc_forw_np = correlate_numpy(a, v, deltat=deltat, normalize=normalize)
# calculate the model curve for cross-correlation
xcc = gac_np[:, 0]
ampcc = np.correlate(a - np.average(a), v - np.average(v), mode="valid")
if normalize:
    ampcc /= len(a) * countrate**2
ycc = ampcc * np.exp(-xcc / taudiff)

# plotting
fig = plt.figure(figsize=(8, 5))
fig.canvas.set_window_title('comparing multipletau')

# autocorrelation
ax1 = fig.add_subplot(211)
ax1.plot(gac_np[:, 0], gac_np[:, 1], "-",
         color="gray", label="correlate (numpy)")
ax1.plot(x, y, "g-", label="input model")
ax1.plot(gac_mt[:, 0], gac_mt[:, 1], "-",
         color="#B60000", label="autocorrelate")
ax1.legend(loc=0, fontsize='small')
ax1.set_xlabel("lag channel")
ax1.set_ylabel("autocorrelation")
ax1.set_xscale('log')
ax1.set_xlim(x.min(), x.max())
ax1.set_ylim(-y.max()*.2, y.max()*1.1)

# cross-correlation
ax2 = fig.add_subplot(212)
ax2.plot(gcc_forw_np[:, 0], gcc_forw_np[:, 1], "-",
         color="gray", label="forward (numpy)")
ax2.plot(xcc, ycc, "g-", label="input model")
ax2.plot(gcc_forw_mt[:, 0], gcc_forw_mt[:, 1], "-",
         color="#B60000", label="forward")
ax2.plot(gcc_back_mt[:, 0], gcc_back_mt[:, 1], "-",
         color="#5D00B6", label="backward")
ax2.set_xlabel("lag channel")
ax2.set_ylabel("cross-correlation")
ax2.legend(loc=0, fontsize='small')
ax2.set_xscale('log')
ax2.set_xlim(x.min(), x.max())
ax2.set_ylim(-ycc.max()*.2, ycc.max()*1.1)

plt.tight_layout()
plt.show()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
A multiple-τ algorithm for Python 2.7 and 3.x.

Copyright (c) 2014 Paul Müller

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

  3. Neither the name of multipletau nor the names of its contributors
     may be used to endorse or promote products derived from this
     software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INFRAE OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from __future__ import division

import numpy as np
import warnings

__all__ = ["autocorrelate", "correlate", "correlate_numpy"]


#: Defines the cutoff when the absolute mean trace divided by the median
#: of the absolute value of the trace is treated as zero.
ZERO_CUTOFF = 1e-15


class DtypeWarning(UserWarning):
    pass


class InvalidMWarning(UserWarning):
    pass


def autocorrelate(a, m=16, deltat=1, normalize=False, copy=True, dtype=None,
                  compress="average", ret_sum=False):
    """
    Autocorrelation of a 1-dimensional sequence on a log2-scale.

    This computes the correlation similar to
    :py:func:`numpy.correlate` for positive :math:`k` on a base 2
    logarithmic scale.

        :func:`numpy.correlate(a, a, mode="full")[len(a)-1:]`

        :math:`z_k = \\Sigma_n a_n a_{n+k}`


    Parameters
    ----------
    a: array-like
        input sequence
    m: even integer
        defines the number of points on one level, must be an
        even integer
    deltat: float
        distance between bins
    normalize: bool
        normalize the result to the square of the average input
        signal and the factor :math:`M-k`.
    copy: bool
        copy input array, set to ``False`` to save memory
    dtype: object to be converted to a data type object
        The data type of the returned array and of the accumulator
        for the multiple-tau computation.
    compress: str
        strategy for propagating values to the next register

        - `"average"` (default): average two measurements when pushing
          to the next level of the correlator.
        - `"first"`: use only the first value when pushing to the next
          level of the correlator.
        - `"second"`: use only the second value when pushing to the
          next level of the correlator.

        Using only the first or the second values during propagation
        completely removes the systematic error at the cost of
        increasing the statistical error.
        See https://doi.org/10.1063/1.3491098 for a discussion on the
        effect of averaging.
    ret_sum: bool
        return the exact sum :math:`z_k = \\Sigma_n a_n a_{n+k}`. In addition
        :math:`M-k` is returned as an array of length N.


    Returns
    -------
    autocorrelation: ndarray of shape (N,2)
        the lag time (1st column) and the autocorrelation (2nd column).
    count: ndarray of length N
        only returned if `ret_sum` is True; the value of :math:`M-k`
        for each row in `autocorrelation`.


    Notes
    -----
    .. versionchanged:: 0.1.6
       Compute the correlation for zero lag time.

    The algorithm computes the correlation with the convention of the
    curve decaying to zero.

    For experiments like e.g. fluorescence correlation spectroscopy,
    the signal can be normalized to :math:`M-k`
    by invoking ``normalize=True``.

    For normalizing according to the behavior
    of :py:func:`numpy.correlate`, use ``normalize=False``.

    For complex arrays, this method falls back to the method
    :func:`correlate`.


    Examples
    --------
    >>> from multipletau import autocorrelate
    >>> autocorrelate(range(42), m=2, dtype=np.float64)
    array([[  0.00000000e+00,   2.38210000e+04],
           [  1.00000000e+00,   2.29600000e+04],
           [  2.00000000e+00,   2.21000000e+04],
           [  4.00000000e+00,   2.03775000e+04],
           [  8.00000000e+00,   1.50612000e+04]])
    """
    if not isinstance(normalize, bool):
        raise ValueError("`normalize` must be boolean!")
    if not isinstance(copy, bool):
        raise ValueError("`copy` must be boolean!")
    if not isinstance(ret_sum, bool):
        raise ValueError("`ret_sum` must be boolean!")
    if normalize and ret_sum:
        raise ValueError("'normalize' and 'ret_sum' must not both be True!")
    compress_values = ["average", "first", "second"]
    if compress not in compress_values:
        raise ValueError("Invalid value for `compress`! Possible values "
                         "are '{}'.".format(','.join(compress_values)))

    if dtype is None:
        dtype = np.dtype(a[0].__class__)
    else:
        dtype = np.dtype(dtype)

    # Complex data
    if dtype.kind == "c":
        # run cross-correlation
        return correlate(a=a,
                         v=a,
                         m=m,
                         deltat=deltat,
                         normalize=normalize,
                         copy=copy,
                         dtype=dtype)
    elif dtype.kind != "f":
        warnings.warn("Input dtype is not float; casting to np.float64!",
                      DtypeWarning)
        dtype = np.dtype(np.float64)

    # If copy is false and dtype is the same as the input array,
    # then this line does not have an effect:
    trace = np.array(a, dtype=dtype, copy=copy)

    # Check parameters
    if m // 2 != m / 2:
        mold = m
        m = np.int_((m // 2 + 1) * 2)
        warnings.warn("Invalid value of m={}. Using m={} instead"
                      .format(mold, m), InvalidMWarning)
    else:
        m = np.int_(m)

    N = N0 = trace.shape[0]

    # Find out the length of the correlation function.
    # The integer k defines how many times we can average over
    # two neighboring array elements in order to obtain an array of
    # length just larger than m.
    k = np.int_(np.floor(np.log2(N / m)))

    # In the base2 multiple-tau scheme, the length of the correlation
    # array is (only taking into account values that are computed from
    # traces that are just larger than m):
    lenG = m + k * (m // 2) + 1

    G = np.zeros((lenG, 2), dtype=dtype)

    normstat = np.zeros(lenG, dtype=dtype)
    normnump = np.zeros(lenG, dtype=dtype)

    traceavg = np.average(trace)

    # We use the fluctuation of the signal around the mean
    if normalize:
        trace -= traceavg
        if np.abs(traceavg) / np.median(np.abs(trace)) < ZERO_CUTOFF:
            raise ValueError("Cannot normalize: Average of `a` is zero!")

    # Otherwise the following for-loop will fail:
    if N < 2 * m:
        raise ValueError("`len(a)` must be >= `2m`!")

    # Calculate autocorrelation function for first m+1 bins
    # Discrete convolution of m elements
    for n in range(0, m + 1):
        G[n, 0] = deltat * n
        # This is the computationally intensive step
        G[n, 1] = np.sum(trace[:N - n] * trace[n:])
        normstat[n] = N - n
        normnump[n] = N
    # Now that we calculated the first m elements of G, let us
    # go on with the next m/2 elements.
    # Check if len(trace) is even:
    if N % 2 == 1:
        N -= 1
    # compress every second element
    if compress == compress_values[0]:
        trace = (trace[:N:2] + trace[1:N:2]) / 2
    elif compress == compress_values[1]:
        trace = trace[:N:2]
    elif compress == compress_values[2]:
        trace = trace[1:N:2]
    N //= 2
    # Start iteration for each m/2 values
    for step in range(1, k + 1):
        # Get the next m/2 values via correlation of the trace
        for n in range(1, m // 2 + 1):
            npmd2 = n + m // 2
            idx = m + n + (step - 1) * m // 2
            if len(trace[:N - npmd2]) == 0:
                # This is a shortcut that stops the iteration once the
                # length of the trace is too small to compute a corre-
                # lation. The actual length of the correlation function
                # does not only depend on k - We also must be able to
                # perform the sum with respect to k for all elements.
                # For small N, the sum over zero elements would be
                # computed here.
                #
                # One could make this for-loop go up to maxval, where
                #   maxval1 = int(m/2)
                #   maxval2 = int(N-m/2-1)
                #   maxval = min(maxval1, maxval2)
                # However, we then would also need to find out which
                # element in G is the last element...
                G = G[:idx - 1]
                normstat = normstat[:idx - 1]
                normnump = normnump[:idx - 1]
                # Note that this break only breaks out of the current
                # for loop. However, we are already in the last loop
                # of the step-for-loop. That is because we calculated
                # k in advance.
                break
            else:
                G[idx, 0] = deltat * npmd2 * 2**step
                # This is the computationally intensive step
                G[idx, 1] = np.sum(trace[:N - npmd2] *
                                   trace[npmd2:])
                normstat[idx] = N - npmd2
                normnump[idx] = N
        # Check if len(trace) is even:
        if N % 2 == 1:
            N -= 1
        # compress every second element
        if compress == compress_values[0]:
            trace = (trace[:N:2] + trace[1:N:2]) / 2
        elif compress == compress_values[1]:
            trace = trace[:N:2]
        elif compress == compress_values[2]:
            trace = trace[1:N:2]

        N //= 2

    if normalize:
        G[:, 1] /= traceavg**2 * normstat
    elif not ret_sum:
        G[:, 1] *= N0 / normnump

    if ret_sum:
        return G, normstat
    else:
        return G


def correlate(a, v, m=16, deltat=1, normalize=False, copy=True, dtype=None,
              compress="average", ret_sum=False):
    """
    Cross-correlation of two 1-dimensional sequences
    on a log2-scale.

    This computes the cross-correlation similar to
    :py:func:`numpy.correlate` for positive :math:`k`  on a base 2
    logarithmic scale.

        :func:`numpy.correlate(a, v, mode="full")[len(a)-1:]`

        :math:`z_k = \\Sigma_n a_n v_{n+k}`

    Note that only the correlation in the positive direction is
    computed. To obtain the correlation for negative lag times
    swap the input variables ``a`` and ``v``.

    Parameters
    ----------
    a, v: array-like
        input sequences with equal length
    m: even integer
        defines the number of points on one level, must be an
        even integer
    deltat: float
        distance between bins
    normalize: bool
        normalize the result to the square of the average input
        signal and the factor :math:`M-k`.
    copy: bool
        copy input array, set to ``False`` to save memory
    dtype: object to be converted to a data type object
        The data type of the returned array and of the accumulator
        for the multiple-tau computation.
    compress: str
        strategy for propagating values to the next register

        - `"average"` (default): average two measurements when pushing
          to the next level of the correlator.
        - `"first"`: use only the first value when pushing to the next
          level of the correlator.
        - `"second"`: use only the second value when pushing to the
          next level of the correlator.

        Using only the first or the second values during propagation
        completely removes the systematic error at the cost of
        increasing the statistical error.
        See https://doi.org/10.1063/1.3491098 for a discussion on the
        effect of averaging.
    ret_sum: bool
        return the exact sum :math:`z_k = \\Sigma_n a_n v_{n+k}`. In addition
        :math:`M-k` is returned as an array of length N.


    Returns
    -------
    cross_correlation: ndarray of shape (N,2)
        the lag time (1st column), the cross-correlation (2nd column).
    count: ndarray of length N
        only returned if `ret_sum` is True; the value of :math:`M-k`
        for each row in `autocorrelation`.


    Notes
    -----
    .. versionchanged:: 0.1.6
       Compute the correlation for zero lag time and correctly
       normalize the correlation for a complex input sequence `v`.

    The algorithm computes the correlation with the convention of the
    curve decaying to zero.

    For experiments like e.g. fluorescence correlation spectroscopy,
    the signal can be normalized to :math:`M-k`
    by invoking ``normalize=True``.

    For normalizing according to the behavior of
    :py:func:`numpy.correlate`, use ``normalize=False``.


    Examples
    --------
    >>> from multipletau import correlate
    >>> correlate(range(42), range(1,43), m=2, dtype=np.float64)
    array([[  0.00000000e+00,   2.46820000e+04],
           [  1.00000000e+00,   2.38210000e+04],
           [  2.00000000e+00,   2.29600000e+04],
           [  4.00000000e+00,   2.12325000e+04],
           [  8.00000000e+00,   1.58508000e+04]])

    """
    if not isinstance(normalize, bool):
        raise ValueError("`normalize` must be boolean!")
    if not isinstance(copy, bool):
        raise ValueError("`copy` must be boolean!")
    if not isinstance(ret_sum, bool):
        raise ValueError("`ret_sum` must be boolean!")
    if normalize and ret_sum:
        raise ValueError("'normalize' and 'ret_sum' must not both be True!")
    compress_values = ["average", "first", "second"]
    if compress not in compress_values:
        raise ValueError("Invalid value for `compress`! Possible values "
                         "are '{}'.".format(','.join(compress_values)))

    # See `autocorrelation` for better documented code.
    traceavg1 = np.average(v)
    traceavg2 = np.average(a)
    if normalize:
        if np.abs(traceavg1) / np.median(np.abs(v)) < ZERO_CUTOFF:
            raise ValueError("Cannot normalize: Average of `v` is zero!")
        if np.abs(traceavg2) / np.median(np.abs(a)) < ZERO_CUTOFF:
            raise ValueError("Cannot normalize: Average of `a` is zero!")

    if dtype is None:
        dtype = np.dtype(v[0].__class__)
        dtype2 = np.dtype(a[0].__class__)
        if dtype != dtype2:
            if dtype.kind == "c" or dtype2.kind == "c":
                # The user might try to combine complex64 and float128.
                warnings.warn(
                    "Input dtypes not equal; casting to np.complex128!",
                    InvalidMWarning)
                dtype = np.dtype(np.complex128)
            else:
                warnings.warn("Input dtypes not equal; casting to np.float64!",
                              InvalidMWarning)
                dtype = np.dtype(np.float64)
    else:
        dtype = np.dtype(dtype)

    if dtype.kind not in ["c", "f"]:
        warnings.warn("Input dtype is not float; casting to np.float64!",
                      InvalidMWarning)
        dtype = np.dtype(np.float64)

    trace1 = np.array(v, dtype=dtype, copy=copy)

    # Prevent traces from overwriting each other
    if a is v:
        # Force copying trace 2
        copy = True

    trace2 = np.array(a, dtype=dtype, copy=copy)

    if trace1.size != trace2.size:
        raise ValueError("`a` and `v` must have same length!")

    # Complex data
    if dtype.kind == "c":
        np.conjugate(trace1, out=trace1)

    # Check parameters
    if m // 2 != m / 2:
        mold = m
        m = np.int_(m // 2 + 1) * 2
        warnings.warn("Invalid value of m={}. Using m={} instead"
                      .format(mold, m), InvalidMWarning)
    else:
        m = np.int_(m)

    N = N0 = trace1.shape[0]
    # Find out the length of the correlation function.
    # The integer k defines how many times we can average over
    # two neighboring array elements in order to obtain an array of
    # length just larger than m.
    k = np.int_(np.floor(np.log2(N / m)))

    # In the base2 multiple-tau scheme, the length of the correlation
    # array is (only taking into account values that are computed from
    # traces that are just larger than m):
    lenG = m + k * m // 2 + 1

    G = np.zeros((lenG, 2), dtype=dtype)
    normstat = np.zeros(lenG, dtype=dtype)
    normnump = np.zeros(lenG, dtype=dtype)

    # We use the fluctuation of the signal around the mean
    if normalize:
        trace1 -= np.conj(traceavg1)
        trace2 -= traceavg2

    # Otherwise the following for-loop will fail:
    if N < 2 * m:
        raise ValueError("`len(a)` must be >= `2m`!")

    # Calculate autocorrelation function for first m+1 bins
    for n in range(0, m + 1):
        G[n, 0] = deltat * n
        G[n, 1] = np.sum(trace1[:N - n] * trace2[n:])
        normstat[n] = N - n
        normnump[n] = N
    # Check if len(trace) is even:
    if N % 2 == 1:
        N -= 1
    # compress every second element
    if compress == compress_values[0]:
        trace1 = (trace1[:N:2] + trace1[1:N:2]) / 2
        trace2 = (trace2[:N:2] + trace2[1:N:2]) / 2
    elif compress == compress_values[1]:
        trace1 = trace1[:N:2]
        trace2 = trace2[:N:2]
    elif compress == compress_values[2]:
        trace1 = trace1[1:N:2]
        trace2 = trace2[1:N:2]
    N //= 2

    for step in range(1, k + 1):
        # Get the next m/2 values of the trace
        for n in range(1, m // 2 + 1):
            npmd2 = (n + m // 2)
            idx = m + n + (step - 1) * m // 2
            if len(trace1[:N - npmd2]) == 0:
                # Abort
                G = G[:idx - 1]
                normstat = normstat[:idx - 1]
                normnump = normnump[:idx - 1]
                break
            else:
                G[idx, 0] = deltat * npmd2 * 2**step
                G[idx, 1] = np.sum(
                    trace1[:N - npmd2] * trace2[npmd2:])
                normstat[idx] = N - npmd2
                normnump[idx] = N

        # Check if len(trace) is even:
        if N % 2 == 1:
            N -= 1
        # compress every second element
        if compress == compress_values[0]:
            trace1 = (trace1[:N:2] + trace1[1:N:2]) / 2
            trace2 = (trace2[:N:2] + trace2[1:N:2]) / 2
        elif compress == compress_values[1]:
            trace1 = trace1[:N:2]
            trace2 = trace2[:N:2]
        elif compress == compress_values[2]:
            trace1 = trace1[1:N:2]
            trace2 = trace2[1:N:2]
        N //= 2

    if normalize:
        G[:, 1] /= traceavg1 * traceavg2 * normstat
    elif not ret_sum:
        G[:, 1] *= N0 / normnump

    if ret_sum:
        return G, normstat
    else:
        return G


def correlate_numpy(a, v, deltat=1, normalize=False,
                    dtype=None, copy=True):
    """
    Convenience function that wraps around :py:func:`numpy.correlate` and
    returns the correlation in the same format as :func:`correlate` does.


    Parameters
    ----------
    a, v: array-like
        input sequences
    deltat: float
        distance between bins
    normalize: bool
        normalize the result to the square of the average input signal
        and the factor :math:`M-k`. The resulting curve follows
        the convention of decaying to zero for large lag times.
    copy: bool
        copy input array, set to ``False`` to save memory
    dtype: object to be converted to a data type object
        The data type of the returned array.


    Returns
    -------
    cross_correlation: ndarray of shape (N,2)
        the lag time (column 1) and the cross-correlation (column 2).


    Notes
    -----
    .. versionchanged:: 0.1.6
       Removed false normalization when `normalize==False`.
    """
    ab = np.array(a, dtype=dtype, copy=copy)
    vb = np.array(v, dtype=dtype, copy=copy)

    if ab.size != vb.size:
        raise ValueError("`a` and `v` must have same length!")

    avg = np.average(ab)
    vvg = np.average(vb)

    if normalize:
        ab -= avg
        vb -= vvg
        if np.abs(avg) / np.median(np.abs(ab)) < ZERO_CUTOFF:
            raise ValueError("Cannot normalize: Average of `a` is zero!")
        if np.abs(vvg) / np.median(np.abs(vb)) < ZERO_CUTOFF:
            raise ValueError("Cannot normalize: Average of `v` is zero!")

    Gd = np.correlate(ab, vb, mode="full")[len(ab) - 1:]

    if normalize:
        N = len(Gd)
        m = N - np.arange(N)
        Gd /= m * avg * vvg

    G = np.zeros((len(Gd), 2), dtype=dtype)
    G[:, 1] = Gd
    G[:, 0] = np.arange(len(Gd)) * deltat
    return G



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests correlation algorithm"""
from __future__ import division, print_function

import sys

import numpy as np
import pytest

import multipletau

from test_autocorrelate import get_reference_data


def get_sample_arrays_cplx():
    a = [-4.3,   1,    9, -99.2, 13]
    b = [9921, 281, 23.5,   5.3, 77]
    c = [12,   0,  2.1,   1.3, 33]
    d = [32,  .1,   -2,   6.3, 88]
    ll = [33,  92,   47,    54, 99]
    r = [0,   1,   12,     4,  0]
    p = [1,   4,   .5,     2,  3]
    arrs = []

    for ai, bi, ci, di, li, ri, pi in zip(a, b, c, d, ll, r, p):
        x = np.linspace(ai, bi, li)
        y = np.linspace(ci, di, li)
        arr = (x*np.roll(x, ri))**pi + 1j*y
        arrs.append(arr)

    return arrs


def test_cc_copy():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays_cplx()

    res1 = []
    for a in arrs:
        r = multipletau.correlate(a=a,
                                  v=a,
                                  m=16,
                                  deltat=1,
                                  normalize=True,
                                  copy=True)
        res1.append(r)

    res2 = []
    for a in arrs:
        r = multipletau.correlate(a=a,
                                  v=a,
                                  m=16,
                                  deltat=1,
                                  normalize=True,
                                  copy=False)
        res2.append(r)

    # simple test if result is the same
    assert np.all(np.concatenate(res1) == np.concatenate(res2))

    arrs = np.concatenate(arrs)
    refarrs = np.concatenate(get_sample_arrays_cplx())

    # make sure the copy function really changes something
    assert not np.all(arrs == refarrs)


@pytest.mark.filterwarnings("ignore::multipletau.core.InvalidMWarning")
def test_cc_dtype():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.round(get_sample_arrays_cplx()[0].real)

    # integer
    rf = multipletau.correlate(a=a,
                               v=a,
                               m=16,
                               deltat=1,
                               normalize=True,
                               copy=True,
                               dtype=np.float64)

    ri = multipletau.correlate(a=a,
                               v=a,
                               m=16,
                               deltat=1,
                               normalize=True,
                               copy=True,
                               dtype=np.int_)

    ri2 = multipletau.correlate(a=np.array(a, dtype=np.int_),
                                v=np.array(a, dtype=np.int_),
                                m=16,
                                deltat=1,
                                normalize=True,
                                copy=True,
                                dtype=None)

    assert ri.dtype == np.dtype(
        np.float64), "if wrong dtype, dtype should default to np.float64"
    assert ri2.dtype == np.dtype(
        np.float64), "if wrong dtype, dtype should default to np.float64"
    assert np.all(
        rf == ri), "result should be the same, because input us the same"
    assert np.all(
        rf == ri2), "result should be the same, because input us the same"


@pytest.mark.filterwarnings("ignore::multipletau.core.InvalidMWarning")
def test_cc_dtype2():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = np.round(get_sample_arrays_cplx()[0])

    rf = multipletau.correlate(a=a.real,
                               v=a,
                               m=16,
                               deltat=1,
                               normalize=True,
                               copy=True)
    assert np.dtype(rf.dtype) == np.dtype(np.complex128)

    rf2 = multipletau.correlate(a=a.real,
                                v=np.array(a.imag, dtype=np.int_),
                                m=16,
                                deltat=1,
                                normalize=True,
                                copy=True)
    assert np.dtype(rf2.dtype) == np.dtype(np.float64)


def test_cc_m():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays_cplx()

    ms = [4, 8, 10, 16, 20, 64, 128]
    a = np.concatenate(arrs)

    res = []
    for m in ms:
        r = multipletau.correlate(a=a,
                                  v=a,
                                  m=m,
                                  deltat=1,
                                  normalize=False,
                                  copy=True,
                                  dtype=np.complex128)
        res.append(r)

        # test minimal length of array
        multipletau.correlate(a=a[:2*m],
                              v=a[:2*m],
                              m=m,
                              deltat=1,
                              normalize=False,
                              copy=True,
                              dtype=np.complex128)

    res = np.concatenate(res)
    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-15)


@pytest.mark.filterwarnings("ignore::multipletau.core.InvalidMWarning")
def test_cc_m_wrong():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    a = get_sample_arrays_cplx()[0]

    # integer
    r1 = multipletau.correlate(a=a,
                               v=a,
                               m=16,
                               deltat=1,
                               normalize=True,
                               copy=True)

    r2 = multipletau.correlate(a=a,
                               v=a,
                               m=15,
                               deltat=1,
                               normalize=True,
                               copy=True)

    r3 = multipletau.correlate(a=a,
                               v=a,
                               m=15.5,
                               deltat=1,
                               normalize=True,
                               copy=True)

    r4 = multipletau.correlate(a=a,
                               v=a,
                               m=14.5,
                               deltat=1,
                               normalize=True,
                               copy=True)

    r5 = multipletau.correlate(a=a,
                               v=a,
                               m=16.,
                               deltat=1,
                               normalize=True,
                               copy=True)

    assert np.all(r1 == r2)
    assert np.all(r1 == r3)
    assert np.all(r1 == r4)
    assert np.all(r1 == r5)


def test_cc_normalize():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays_cplx()

    res = []
    for a in arrs:
        r = multipletau.correlate(a=a.real,
                                  v=a.imag,
                                  m=16,
                                  deltat=1,
                                  normalize=True,
                                  copy=True,
                                  dtype=np.float64)
        res.append(r)
    res = np.concatenate(res)
    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-14)


def test_cc_simple():
    myframe = sys._getframe()
    myname = myframe.f_code.co_name
    print("running ", myname)

    arrs = get_sample_arrays_cplx()

    res = []
    for a in arrs:
        r = multipletau.correlate(a=a,
                                  v=a,
                                  m=16,
                                  deltat=1,
                                  normalize=False,
                                  copy=True,
                                  dtype=np.complex128)
        res.append(r)
    res = np.concatenate(res)

    # np.save(os.path.dirname(__file__)
    #         + "/data/"+os.path.basename(__file__)+"_"+myname+".npy", res)
    ref = get_reference_data(myname, __file__)

    assert np.allclose(res, ref, atol=0, rtol=1e-15)

    # also check result of autocorrelate
    res2 = []
    for a in arrs:
        r = multipletau.autocorrelate(a=a,
                                      m=16,
                                      deltat=1,
                                      normalize=False,
                                      copy=True,
                                      dtype=np.complex128)
        res2.append(r)
    res2 = np.concatenate(res2)

    assert np.allclose(res, res2, atol=0, rtol=1e-15)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
